/*
	Linux Doom Editor, by Peter Amstutz

	based on the DOS program:
   Amazing Doom Editor, by Brendon Wyber and Raphal Quinet.
	(which in turn was originally the famous DEU)

   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.

   This program comes with absolutely no warranty.

   GFX.C - Graphics routines.  Now uses GGI!
*/

/* the includes */
#include "deu.h"
#include <math.h>

/* the global variables */
/*
  int GfxMode = 0;         graphics mode number, or 0 for text 
  1 = 320x200, 2 = 640x480, 3 = 800x600, 4 = 1024x768 
  positive = 16 colors, negative = 256 colors 
  */

float Scale;            /* the scale value */
int PointerX;           /* X position of pointer */
int PointerY;           /* Y position of pointer */
int ScrMaxX;            /* maximum X screen coord */
int ScrMaxY;            /* maximum Y screen coord */
int ScrCenterX;         /* X coord of screen center */
int ScrCenterY;         /* Y coord of screen center */

extern int drewsomething;

ggi_visual_t vis=NULL;
ggi_visual_t memvis=NULL;
ggi_uint BLACK, BLUE, GREEN, CYAN, RED, MAGENTA, BROWN, LIGHTGRAY, DARKGRAY,
     LIGHTBLUE, LIGHTGREEN, LIGHTCYAN, LIGHTRED, LIGHTMAGENTA, YELLOW, WHITE;
int xres=640, yres=480;

void SetupColors()
{
     ggi_color col;
     col.r=0; col.g=0; col.b=0;
     BLACK=ggiMapColor(vis, &col);
     col.r=0; col.g=0; col.b=0xFFFF/2;
     BLUE=ggiMapColor(vis, &col);
     col.r=0; col.g=0xFFFF/2; col.b=0;
     GREEN=ggiMapColor(vis, &col);
     col.r=0; col.g=0xFFFF/2; col.b=0xFFFF/2;
     CYAN=ggiMapColor(vis, &col);
     col.r=0xFFFF/2; col.g=0; col.b=0;
     RED=ggiMapColor(vis, &col);
     col.r=0xFFFF/2; col.g=0; col.b=0xFFFF/2;
     MAGENTA=ggiMapColor(vis, &col);
     col.r=0xFFFF/2; col.g=0xFFFF/2; col.b=0;
     BROWN=ggiMapColor(vis, &col);
     col.r=(0xFFFF*2)/3; col.g=(0xFFFF*2)/3; col.b=(0xFFFF*2)/3;
     LIGHTGRAY=ggiMapColor(vis, &col);
     col.r=0xFFFF/3; col.g=0xFFFF/3; col.b=0xFFFF/3;
     DARKGRAY=ggiMapColor(vis, &col);
     col.r=0; col.g=0; col.b=0xFFFF;
     LIGHTBLUE=ggiMapColor(vis, &col);
     col.r=0; col.g=0xFFFF; col.b=0;
     LIGHTGREEN=ggiMapColor(vis, &col);
     col.r=0; col.g=0xFFFF; col.b=0xFFFF;
     LIGHTCYAN=ggiMapColor(vis, &col);
     col.r=0xFFFF; col.g=0; col.b=0;
     LIGHTRED=ggiMapColor(vis, &col);
     col.r=0xFFFF; col.g=0; col.b=0xFFFF;
     LIGHTMAGENTA=ggiMapColor(vis, &col);
     col.r=0xFFFF; col.g=0xFFFF; col.b=0;
     YELLOW=ggiMapColor(vis, &col);
     col.r=0xFFFF; col.g=0xFFFF; col.b=0xFFFF;
     WHITE=ggiMapColor(vis, &col);
}


/*
   initialise the graphics display
*/

void InitGfx()
{
     ggi_graphtype modes[6] = { GT_8BIT, GT_15BIT, GT_16BIT, GT_24BIT,
				GT_32BIT, GT_4BIT };
     ggi_mode mode;
     int i;

     if(ggiInit()!=0)
     {
	  ProgError("Error initializing GGI");
     }
     
     vis=ggiOpen(NULL);
     mode.frames=1;
     mode.visible.x=xres;
     mode.visible.y=yres;
     mode.virt.x=GGI_AUTO;
     mode.virt.y=GGI_AUTO;
	 /*
     mode.text.x=8;
     mode.text.y=8;
	 */

     ScrMaxX=mode.visible.x-1;
     ScrMaxY=mode.visible.y-1;
     ScrCenterX=ScrMaxX/2;
     ScrCenterY=ScrMaxY/2;

     for(i=0; i<6; i++)
     {
	  mode.graphtype=modes[i];
	  if(ggiCheckMode(vis, &mode)==0)
	  {
	       if(i==5)
	       {
		    if(ggiCheckGraphMode(vis, 320, 200, 320, 200, GT_8BIT, NULL)!=0)
		    {
			 ProgError("Error: can't even support 320x200x256!");
		    }
	       }
	       ggiSetMode(vis, &mode);
	       if(i==0)
	       {
		    SetDoomPalette(vis, 0);
	       }
	       break;
	  }
     }
     if(i==6) 
     {
	  ProgError("Can't find any support graphics modes!");
     }

     SetupColors();
     memvis=ggiOpen("display-memory",NULL);
     ggiSetGraphMode(memvis, 400, 400, 400, 400, GT_8BIT);
     SetDoomPalette(memvis, 0);

     ggiSetGCBackground(vis, BLACK);
	 ggiSetInfoFlags(vis, GGIFLAG_ASYNC);
	 ggiFlush(vis);
}



/*
   terminate the graphics display
*/

void TermGfx()
{
     if(vis) 
     {
	  ggiClose(vis);
	  ggiClose(memvis);
	  vis=NULL;
	  memvis=NULL;
	  ggiExit();
     }
}



/*
   switch from VGA 16 colours to VGA 256 colours
*/

Bool SwitchToVGA256()
{

     ggi_mode mode;

     mode.frames=1;
     mode.visible.x=320;
     mode.visible.y=200;
     mode.virt.x=320;
     mode.virt.y=200;
	 /*
     mode.text.x=8;
     mode.text.y=8;
	 */
     mode.graphtype=GT_8BIT;

     ScrMaxX=mode.visible.x;
     ScrMaxY=mode.visible.y;
     ScrCenterX=ScrMaxX/2;
     ScrCenterY=ScrMaxY/2;

     ggiSetMode(vis, &mode);
     SetDoomPalette(vis, 0);     
}



/*
   switch from VGA 256 colours to VGA 16 colours
*/

Bool SwitchToVGA16()
{

     ggi_mode mode;

     mode.frames=1;
     mode.visible.x=640;
     mode.visible.y=480;
     mode.virt.x=640;
     mode.virt.y=480;
	 /*
     mode.text.x=8;
     mode.text.y=8;
	 */
     mode.graphtype=GT_4BIT;

     ScrMaxX=mode.visible.x;
     ScrMaxY=mode.visible.y;
     ScrCenterX=ScrMaxX/2;
     ScrCenterY=ScrMaxY/2;

     ggiSetMode(vis, &mode);
}



/*
   clear the screen
*/

void ClearScreen()
{
     SetColor(BLACK);
     ggiFillscreen(vis);
}



/*
   set the current drawing color
*/

void SetColor( int color)
{
     ggiSetGCForeground(vis, color);
}



/*
   draw a line on the screen from map coords
*/

void DrawMapLine( int mapXstart, int mapYstart, int mapXend, int mapYend)
{
     ggiDrawLine(vis, SCREENX( mapXstart), SCREENY( mapYstart), SCREENX( mapXend), SCREENY( mapYend));
	 drewsomething=1;
}



/*
   draw a circle on the screen from map coords
*/

void DrawMapCircle( int mapXcenter, int mapYcenter, int mapRadius)
{
     ggiDrawCircle(vis, SCREENX( mapXcenter), SCREENY( mapYcenter), (int) (mapRadius * Scale));
	 drewsomething=1;
}



/*
   draw an arrow on the screen from map coords
*/

void DrawMapVector( int mapXstart, int mapYstart, int mapXend, int mapYend)
{
     int    scrXstart = SCREENX( mapXstart);
     int    scrYstart = SCREENY( mapYstart);
     int    scrXend   = SCREENX( mapXend);
     int    scrYend   = SCREENY( mapYend);
     double r         = hypot( (double) (scrXstart - scrXend), (double) (scrYstart - scrYend));
     int    scrXoff   = (r >= 1.0) ? (int) ((scrXstart - scrXend) * 8.0 / r * Scale) : 0;
     int    scrYoff   = (r >= 1.0) ? (int) ((scrYstart - scrYend) * 8.0 / r * Scale) : 0;

     ggiDrawLine(vis, scrXstart, scrYstart, scrXend, scrYend);
     scrXstart = scrXend + 2 * scrXoff;
     scrYstart = scrYend + 2 * scrYoff;
     ggiDrawLine(vis, scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend);
     ggiDrawLine(vis, scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend);
/*
  line( scrXstart - scrYoff, scrYstart + scrXoff, scrXend + scrXoff, scrYend + scrYoff);
  line( scrXstart + scrYoff, scrYstart - scrXoff, scrXend + scrXoff, scrYend + scrYoff);
  */
  drewsomething=1;
}



/*
   draw an arrow on the screen from map coords and angle (0 - 65535)
*/

void DrawMapArrow( int mapXstart, int mapYstart, unsigned angle)
{
     int    mapXend   = mapXstart + (int) (50 * cos(angle / 10430.37835));
     int    mapYend   = mapYstart + (int) (50 * sin(angle / 10430.37835));
     int    scrXstart = SCREENX( mapXstart);
     int    scrYstart = SCREENY( mapYstart);
     int    scrXend   = SCREENX( mapXend);
     int    scrYend   = SCREENY( mapYend);
     double r         = hypot( scrXstart - scrXend, scrYstart - scrYend);
     int    scrXoff   = (r >= 1.0) ? (int) ((scrXstart - scrXend) * 8.0 / r * Scale) : 0;
     int    scrYoff   = (r >= 1.0) ? (int) ((scrYstart - scrYend) * 8.0 / r * Scale) : 0;

     ggiDrawLine(vis, scrXstart, scrYstart, scrXend, scrYend);
     scrXstart = scrXend + 2 * scrXoff;
     scrYstart = scrYend + 2 * scrYoff;
     ggiDrawLine(vis, scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend);
     ggiDrawLine(vis, scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend);
	 drewsomething=1;
}



/*
   draw a line on the screen from screen coords
*/

void DrawScreenLine( int Xstart, int Ystart, int Xend, int Yend)
{
     ggiDrawLine(vis, Xstart, Ystart, Xend, Yend);
	 drewsomething=1;
}



/*
   draw a filled in box on the screen from screen coords
*/

void DrawScreenBox( int Xstart, int Ystart, int Xend, int Yend)
{
     ggiDrawBox(vis, Xstart, Ystart, Xend-Xstart, Yend-Ystart);
	 drewsomething=1;
}


/*
   draw a filled-in 3D-box on the screen from screen coords
*/

void DrawScreenBox3D( int Xstart, int Ystart, int Xend, int Yend)
{
     /*    setfillstyle( 1, TranslateToDoomColor( LIGHTGRAY)); */
     ggiSetGCBackground(vis, LIGHTGRAY);
     ggiSetGCForeground(vis, LIGHTGRAY);
     DrawScreenBox(Xstart + 1, Ystart + 1, Xend - 1, Yend - 1);
     ggiSetGCForeground(vis, DARKGRAY);
     ggiDrawLine(vis, Xstart, Yend, Xend, Yend);
     ggiDrawLine(vis, Xend, Ystart, Xend, Yend);
     if (Xend - Xstart > 20 && Yend - Ystart > 20)
     {
	  ggiDrawLine(vis, Xstart + 1, Yend - 1, Xend - 1, Yend - 1);
	  ggiDrawLine(vis, Xend - 1, Ystart + 1, Xend - 1, Yend - 1);
	  SetColor(WHITE);
	  ggiDrawLine(vis, Xstart + 1, Ystart + 1, Xstart + 1, Yend - 1);
	  ggiDrawLine(vis, Xstart + 1, Ystart + 1, Xend - 1, Ystart + 1);
     }
     SetColor( WHITE);
     ggiDrawLine(vis, Xstart, Ystart, Xend, Ystart);
     ggiDrawLine(vis, Xstart, Ystart, Xstart, Yend);
     SetColor(BLACK);
	 drewsomething=1;
}



/*
   draw a hollow 3D-box on the screen from screen coords
*/

void DrawScreenBoxHollow( int Xstart, int Ystart, int Xend, int Yend)
{
     SetColor(BLACK);
     DrawScreenBox(Xstart + 1, Ystart + 1, Xend - 1, Yend - 1);
     SetColor(WHITE);
     ggiDrawLine(vis, Xstart, Yend, Xend, Yend);
     ggiDrawLine(vis, Xend, Ystart, Xend, Yend);
     if (Xend - Xstart > 20 && Yend - Ystart > 20)
     {
	  ggiDrawLine(vis, Xstart + 1, Yend - 1, Xend - 1, Yend - 1);
	  ggiDrawLine(vis, Xend - 1, Ystart + 1, Xend - 1, Yend - 1);
	  SetColor(DARKGRAY);
	  ggiDrawLine(vis, Xstart + 1, Ystart + 1, Xstart + 1, Yend - 1);
	  ggiDrawLine(vis, Xstart + 1, Ystart + 1, Xend - 1, Ystart + 1);
     }
     SetColor( DARKGRAY);
     ggiDrawLine(vis, Xstart, Ystart, Xend, Ystart);
     ggiDrawLine(vis, Xstart, Ystart, Xstart, Yend);
     SetColor( WHITE);
	 drewsomething=1;
}



/*
   draw a meter bar on the screen from screen coords (in a hollow box); max. value = 1.0
*/

void DrawScreenMeter( int Xstart, int Ystart, int Xend, int Yend, float value)
{
     if (value < 0.0)
	  value = 0.0;
     if (value > 1.0)
	  value = 1.0;
     SetColor(BLACK);
     ggiDrawBox(vis,  Xstart + 1 + (int) ((Xend - Xstart - 2) * value), Ystart + 1, Xend - 1, Yend - 1);
     SetColor(LIGHTGREEN);
     ggiDrawBox(vis,  Xstart + 1, Ystart + 1, Xstart + 1 + (int) ((Xend - Xstart - 2) * value), Yend - 1);
	 drewsomething=1;
}



/*
   write text to the screen
*/

void DrawScreenText( int Xstart, int Ystart, char *msg, ...)
{
     static int lastX;
     static int lastY;
     char temp[ 120];
     va_list args;

     va_start( args, msg);
     vsprintf( temp, msg, args);
     va_end( args);
     if (Xstart < 0)
	  Xstart = lastX;
     if (Ystart < 0)
	  Ystart = lastY;
/*     outtextxy( Xstart, Ystart, temp); */
     ggiPuts(vis, Xstart, Ystart, temp);
     lastX = Xstart;
     lastY = Ystart + 10;  /* or textheight("W") ? */
	 drewsomething=1;
}


char erasebuf[3600];

void UpdatePointer()
{
	  ggiGetBox(vis, PointerX - 15, PointerY - 13, 31, 27, erasebuf);
}

/*
   draw the pointer if we aren't using the mouse
*/

int oldptrx=0, oldptry=0;
void DrawPointer( Bool rulers)
{
     int r;
/*	if(PointerX==oldptrx && PointerY==oldptry) return;
	oldptrx=PointerX; oldptry=PointerY; */

	if(!FakeCursor) return;
     /* use XOR mode : drawing the pointer twice erases it */
/*     setwritemode( XOR_PUT); */
     /* draw the pointer */
     if ( rulers)
     {
	  SetColor( MAGENTA);
	  r = (int) (512 * Scale);
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r >>= 1;
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r >>= 1;
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r >>= 1;
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r = (int) (1024 * Scale);
	  ggiDrawLine(vis, PointerX - r, PointerY, PointerX + r, PointerY);
	  ggiDrawLine(vis, PointerX, PointerY - r, PointerX, PointerY + r);
     }
     else
     {
	  UpdatePointer();
	  SetColor( YELLOW);
	  ggiDrawLine(vis, PointerX - 15, PointerY - 13, PointerX + 15, PointerY + 13);
	  ggiDrawLine(vis, PointerX - 15, PointerY + 13, PointerX + 15, PointerY - 13);
     }
     /* restore normal write mode */
/*     setwritemode( COPY_PUT); */
	drewsomething=1;
}

/* Erase that little yellow pointer thingy */


void ErasePointer( Bool rulers)
{

     int r;

/*	if(PointerX==oldptrx && PointerY==oldptry) return;
	oldptrx=PointerX; oldptry=PointerY; */
	if(!FakeCursor) return;
     /* use XOR mode : drawing the pointer twice erases it */
/*     setwritemode( XOR_PUT); */
     /* draw the pointer */
     if ( rulers)
     {
	  SetColor( BLACK);
	  r = (int) (512 * Scale);
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r >>= 1;
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r >>= 1;
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r >>= 1;
	  ggiDrawCircle(vis, PointerX, PointerY, r);
	  r = (int) (1024 * Scale);
	  ggiDrawLine(vis, PointerX - r, PointerY, PointerX + r, PointerY);
	  ggiDrawLine(vis, PointerX, PointerY - r, PointerX, PointerY + r);
     }
     else
     {
	  ggiPutBox(vis, PointerX - 15, PointerY - 13, 31, 27, erasebuf);
     }
     /* restore normal write mode */
/*     setwritemode( COPY_PUT); */
	drewsomething=1;
}



/*
   load one "playpal" palette and change all palette colours
*/

void SetDoomPalette(ggi_visual_t v,  int playpalnum)
{
     MDirPtr             dir;
     unsigned char  dpal[768];
     ggi_color cmap[256];
     int                 n;

     if (playpalnum < 0 && playpalnum > 13)
	  return;
     dir = FindMasterDir( MasterDir, "PLAYPAL");
     if (dir)
     {
	  BasicWadSeek( dir->wadfile, dir->dir.start);
	  for (n = 0; n <= playpalnum; n++)
	       BasicWadRead( dir->wadfile, dpal, 768L);
	  for(n=0; n<256; n++)
	  {

	       cmap[n].r=dpal[n*3]<<8;
	       cmap[n].g=dpal[n*3+1]<<8;
	       cmap[n].b=dpal[n*3+2]<<8;
	  }    
	  ggiSetPaletteVec(v, 0, 256, cmap);
     }
}


/*
   translate (dx, dy) into an integer angle value (0-65535)
*/

unsigned ComputeAngle( int dx, int dy)
{
     return (unsigned) (atan2( (double) dy, (double) dx) * 10430.37835 + 0.5);
     /* Yes, I know this function could be in another file, but */
     /* this is the only source file that includes <math.h>...  */
}



/*
   compute the distance from (0, 0) to (dx, dy)
*/

unsigned ComputeDist( int dx, int dy)
{
     return (unsigned) (hypot( (double) dx, (double) dy) + 0.5);
     /* Yes, I know this function could be in another file, but */
     /* this is the only source file that includes <math.h>...  */
}



/*
   insert the vertices of a new polygon
*/

void InsertPolygonVertices( int centerx, int centery, int sides, int radius)
{
     int n;

     for (n = 0; n < sides; n++)
	  InsertObject( OBJ_VERTEXES, 
			-1, 
			centerx + (int) ((double) radius * cos( 6.28 * (double) n / (double) sides)), 
			centery + (int) ((double) radius * sin( 6.2832 * (double) n / (double) sides)));
     /* Yes, I know... etc. */
}



/*
   move (x, y) to a new position: rotate and scale around (0, 0)
*/

void RotateAndScaleCoords( int *x, int *y, double angle, double scale)
{
     double r, theta;

     r = hypot( (double) *x, (double) *y);
     theta = atan2( (double) *y, (double) *x);
     *x = (int) (r * scale * cos( theta + angle) + 0.5);
     *y = (int) (r * scale * sin( theta + angle) + 0.5);
     /* Yes, I know... etc. */
}

/* end of file */
